/*
 * IGMP.c
 *
 * Created: 31.12.2018 19:23:11
 *  Author: Thomas
 */ 


#include <stdbool.h>
#include <string.h>

#include "Internet/Internet.h"
#include "Internet/IP/IPV4.h"
#include "IGMP.h"



int IgmpReadFrame(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader, IGMP_FRAME_Type* igmpFrame)
{
	int ret;
	
	ret = IgmpReadFrameV1(buffer, size, ipHeader, igmpFrame);
	if (ret >= 0) { return 1;}
	
	ret = IgmpReadFrameV2(buffer, size, ipHeader, igmpFrame);
	if (ret >= 0) { return 2;}
	
	return -1;
}




int IgmpReadFrameV1(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader, IGMP_FRAME_Type* igmpFrame)
{
	igmpFrame->IgmpVersion	= 0;

	if (ipHeader->Protocol != IGMP_IPTYPE)
		return -1;
	if (*size < IGMP_HEADER_SIZE_V1)
		return -2;
	if (InetCalcChecksum(*buffer, IGMP_HEADER_SIZE_V1) != 0x0000)
		return -3;

	uint8_t* buf = *buffer;
	
	igmpFrame->Type					= buf[0];
	igmpFrame->Code					= buf[1];
	igmpFrame->Checksum				= (buf[2] << 8) | buf[3];
	igmpFrame->GroupAddr.BYTE[0]	= buf[4];
	igmpFrame->GroupAddr.BYTE[1]	= buf[5];
	igmpFrame->GroupAddr.BYTE[2]	= buf[6];
	igmpFrame->GroupAddr.BYTE[3]	= buf[7];

	igmpFrame->Length = IGMP_HEADER_SIZE_V1;
	
	if (igmpFrame->Code != 0)
		return -4;
	if ((igmpFrame->Type != IGMP_MEMBERSHIP_QUERY) && (igmpFrame->Type != IGMP_V1_MEMBERSHIP_REPORT))
		return -4;

	if (igmpFrame->Type == IGMP_MEMBERSHIP_QUERY)
		if (!IpV4IpEqual(&ipHeader->DestIpAddr, &IGMP_IP_QUERY))
			return -5;

	igmpFrame->IgmpVersion	= 1;

	*buffer	+= IGMP_HEADER_SIZE_V1;
	*size	-= IGMP_HEADER_SIZE_V1;
	
	return 0;	
}



int IgmpReadFrameV2(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader, IGMP_FRAME_Type* igmpFrame)
{
	igmpFrame->IgmpVersion	= 0;

	if (ipHeader->Protocol != IGMP_IPTYPE)
		return -1;
	if (*size < IGMP_HEADER_SIZE_V2)
		return -2;
	if (InetCalcChecksum(*buffer, IGMP_HEADER_SIZE_V2) != 0x0000)
		return -3;

	uint8_t* buf = *buffer;
	
	igmpFrame->Type					= buf[0];
	igmpFrame->Code					= buf[1];
	igmpFrame->Checksum				= (buf[2] << 8) | buf[3];
	igmpFrame->GroupAddr.BYTE[0]	= buf[4];
	igmpFrame->GroupAddr.BYTE[1]	= buf[5];
	igmpFrame->GroupAddr.BYTE[2]	= buf[6];
	igmpFrame->GroupAddr.BYTE[3]	= buf[7];

	igmpFrame->Length = IGMP_HEADER_SIZE_V2;
	
	if ((igmpFrame->Type == IGMP_MEMBERSHIP_QUERY) && (igmpFrame->Code == 0))
		return -4;
	if (	(igmpFrame->Type != IGMP_MEMBERSHIP_QUERY)
		 && (igmpFrame->Type != IGMP_V2_MEMBERSHIP_REPORT)
 		 && (igmpFrame->Type != IGMP_V2_LEAVE_GROUP) )
	{
		return -4;
	}

	if (igmpFrame->Type == IGMP_MEMBERSHIP_QUERY)
		if (!IpV4IpEqual(&ipHeader->DestIpAddr, &IGMP_IP_QUERY))
			return -5;

	igmpFrame->IgmpVersion	= 2;

	*buffer	+= IGMP_HEADER_SIZE_V2;
	*size	-= IGMP_HEADER_SIZE_V2;
	
	return 0;
}



int IgmpWriteFrame(uint8_t* buffer[], uint16_t* size, IGMP_FRAME_Type* igmpFrame)
{
	if (igmpFrame->IgmpVersion == 1)
		return IgmpWriteFrameV1(buffer, size, igmpFrame);

	if (igmpFrame->IgmpVersion == 2)
		return IgmpWriteFrameV2(buffer, size, igmpFrame);
	
	return -1;	
}


int IgmpWriteFrameV1(uint8_t* buffer[], uint16_t* size, IGMP_FRAME_Type* igmpFrame)
{
	if (igmpFrame->IgmpVersion != 1)
		return -1;
	
	if (*size < IGMP_HEADER_SIZE_V1)
		return -2;

	uint8_t* buf = *buffer;

	buf[0] = igmpFrame->Type;
	buf[1] = igmpFrame->Code;
	buf[2] = 0; // Checksum clear
	buf[3] = 0; // Checksum clear
	buf[4] = igmpFrame->GroupAddr.BYTE[0];
	buf[5] = igmpFrame->GroupAddr.BYTE[1];
	buf[6] = igmpFrame->GroupAddr.BYTE[2];
	buf[7] = igmpFrame->GroupAddr.BYTE[3];


	igmpFrame->Checksum = InetCalcChecksum(buf, IGMP_HEADER_SIZE_V1);

	buf[2] = (uint8_t)(igmpFrame->Checksum >> 8);
	buf[3] = (uint8_t)(igmpFrame->Checksum >> 0);
	

	*buffer	+= IGMP_HEADER_SIZE_V1;
	*size	-= IGMP_HEADER_SIZE_V1;
	
	return 0;
}



int IgmpWriteFrameV2(uint8_t* buffer[], uint16_t* size, IGMP_FRAME_Type* igmpFrame)
{
	if (igmpFrame->IgmpVersion != 2)
		return -1;
	
	if (*size < IGMP_HEADER_SIZE_V2)
		return -2;

	uint8_t* buf = *buffer;

	buf[0] = igmpFrame->Type;
	buf[1] = igmpFrame->Code;
	buf[2] = 0; // Checksum clear
	buf[3] = 0; // Checksum clear
	buf[4] = igmpFrame->GroupAddr.BYTE[0];
	buf[5] = igmpFrame->GroupAddr.BYTE[1];
	buf[6] = igmpFrame->GroupAddr.BYTE[2];
	buf[7] = igmpFrame->GroupAddr.BYTE[3];


	igmpFrame->Checksum = InetCalcChecksum(buf, IGMP_HEADER_SIZE_V2);

	buf[2] = (uint8_t)(igmpFrame->Checksum >> 8);
	buf[3] = (uint8_t)(igmpFrame->Checksum >> 0);
	

	*buffer	+= IGMP_HEADER_SIZE_V2;
	*size	-= IGMP_HEADER_SIZE_V2;
	
	return 0;
}



int IgmpSetQueryFrameV1(IGMP_FRAME_Type* igmpFrame)
{
	igmpFrame->Type			= IGMP_MEMBERSHIP_QUERY;
	igmpFrame->Code			= 0;	// Indicates V1
	igmpFrame->Checksum		= 0;
	igmpFrame->GroupAddr	= IP_ANY;

	igmpFrame->IgmpVersion	= 1;
	igmpFrame->Length		= IGMP_HEADER_SIZE_V1;

	return 0;	
}



int IgmpSetQueryFrameV2(uint8_t maxRespTime, IGMP_FRAME_Type* igmpFrame)
{
	if (maxRespTime == 0) { maxRespTime = 10; }

	igmpFrame->Type			= IGMP_MEMBERSHIP_QUERY;
	igmpFrame->Code			= maxRespTime;	// Indicates V2
	igmpFrame->Checksum		= 0;
	igmpFrame->GroupAddr	= IP_ANY;

	igmpFrame->IgmpVersion	= 2;
	igmpFrame->Length		= IGMP_HEADER_SIZE_V2;

	return 0;
}



int IgmpSetReportFrameV1(INET_IP_Type* groupAddr, IGMP_FRAME_Type* igmpFrame)
{
	igmpFrame->Type			= IGMP_V1_MEMBERSHIP_REPORT;
	igmpFrame->Code			= 0;
	igmpFrame->Checksum		= 0;
	igmpFrame->GroupAddr	= *groupAddr;

	igmpFrame->IgmpVersion	= 1;
	igmpFrame->Length		= IGMP_HEADER_SIZE_V1;

	return 0;
}



int IgmpSetReportFrameV2(INET_IP_Type* groupAddr, IGMP_FRAME_Type* igmpFrame)
{
	igmpFrame->Type			= IGMP_V2_MEMBERSHIP_REPORT;
	igmpFrame->Code			= 0;
	igmpFrame->Checksum		= 0;
	igmpFrame->GroupAddr	= *groupAddr;

	igmpFrame->IgmpVersion	= 2;
	igmpFrame->Length		= IGMP_HEADER_SIZE_V2;

	return 0;
}
