/*
 * IPV4.c
 *
 * Created: 27.12.2018 15:28:29
 *  Author: Thomas
 */ 


#include <stdbool.h>
#include <string.h>

#include "Internet/Internet.h"
#include "Internet/ICMP/ICMP.h"
#include "Internet/IGMP/IGMP.h"
#include "IPV4.h"



static inline __attribute__((always_inline)) uint16_t GetNewIpHeaderIdentificationValue(void)
{
	static uint16_t IpHeaderIdentification = 100;
	
	IpHeaderIdentification += 1;
	
	return IpHeaderIdentification;
}



IPV4_PTR_OPTION_Type* IpV4GetRouterAlertOptionPtr(void)
{
	static uint8_t option[] = { 0x94, 0x04, 0x00, 0x00 };
	static IPV4_PTR_OPTION_Type optionPtr;
	
	optionPtr.Buffer = option;
	optionPtr.Length = sizeof(option);
	
	return &optionPtr;
}



int IpV4ReadFrameHeader(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader)
{
	if (*size < IPV4_MIN_HEADER_SIZE)
		return -1;
	
	uint8_t* buf = *buffer;
	
	ipHeader->Version				= buf[0] >> 4;
	ipHeader->HeaderLengthX32B		= buf[0] & 0x0f;
	ipHeader->TOS.BYTE[0]			= buf[1];
	ipHeader->TotalLength			= (buf[2] << 8) | buf[3];
	ipHeader->Identification		= (buf[4] << 8) | buf[5];
	ipHeader->FRAG.hw				= (buf[6] << 8) | buf[7];
	ipHeader->TimeToLive			= buf[8];
	ipHeader->Protocol				= buf[9];
	ipHeader->Checksum				= (buf[10] << 8) | buf[11];
	ipHeader->SrcIpAddr.BYTE[0]		= buf[12]; 
	ipHeader->SrcIpAddr.BYTE[1]		= buf[13];
	ipHeader->SrcIpAddr.BYTE[2]		= buf[14];
	ipHeader->SrcIpAddr.BYTE[3]		= buf[15];
	ipHeader->DestIpAddr.BYTE[0]	= buf[16];
	ipHeader->DestIpAddr.BYTE[1]	= buf[17];
	ipHeader->DestIpAddr.BYTE[2]	= buf[18];
	ipHeader->DestIpAddr.BYTE[3]	= buf[19];

	
	uint16_t headerLength = (ipHeader->HeaderLengthX32B << 2);

	if (*size < headerLength)
		return -2;

	
	if (InetCalcChecksum(buf, headerLength) != 0x0000)
		return -3;
	if (ipHeader->Version != IPV4_VERSION)
		return -4;
	if (ipHeader->FRAG.b.More)
		return -5;
	if (*size < ipHeader->TotalLength)
		return -6;


	ipHeader->PayloadLength = ipHeader->TotalLength - headerLength;
	ipHeader->Option.Length = headerLength - IPV4_MIN_HEADER_SIZE;


	if (ipHeader->Option.Length)
		ipHeader->Option.Buffer = *buffer + IPV4_MIN_HEADER_SIZE;
	else
		ipHeader->Option.Buffer = 0; 
	

	*buffer	+= headerLength;
	*size	-= headerLength;
	
	return 0;
}



int IpV4WriteFrameHeader(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader)
{
	uint16_t headerLength = ipHeader->HeaderLengthX32B << 2;
	
	if (*size < headerLength)
		return -1;

	uint8_t* buf = *buffer;

	buf[0] = (ipHeader->Version << 4) | (ipHeader->HeaderLengthX32B);
	buf[1] = ipHeader->TOS.BYTE[0];
	buf[2] = (uint8_t)(ipHeader->TotalLength >> 8);
	buf[3] = (uint8_t)(ipHeader->TotalLength >> 0);
	buf[4] = (uint8_t)(ipHeader->Identification >> 8);
	buf[5] = (uint8_t)(ipHeader->Identification >> 0);
	buf[6] = (uint8_t)(ipHeader->FRAG.hw >> 8);
	buf[7] = (uint8_t)(ipHeader->FRAG.hw >> 0);
	buf[8] = ipHeader->TimeToLive;
	buf[9] = ipHeader->Protocol;
	buf[10] = 0; // Checksum clear 
	buf[11] = 0; // Checksum clear
	buf[12] = ipHeader->SrcIpAddr.BYTE[0];
	buf[13] = ipHeader->SrcIpAddr.BYTE[1];
	buf[14] = ipHeader->SrcIpAddr.BYTE[2];
	buf[15] = ipHeader->SrcIpAddr.BYTE[3];
	buf[16] = ipHeader->DestIpAddr.BYTE[0];
	buf[17] = ipHeader->DestIpAddr.BYTE[1];
	buf[18] = ipHeader->DestIpAddr.BYTE[2];
	buf[19] = ipHeader->DestIpAddr.BYTE[3];

	if (ipHeader->Option.Length)
		memcpy(&buf[IPV4_MIN_HEADER_SIZE], ipHeader->Option.Buffer, ipHeader->Option.Length);	


	ipHeader->Checksum = InetCalcChecksum(buf, headerLength);

	buf[10] = (uint8_t)(ipHeader->Checksum >> 8);
	buf[11] = (uint8_t)(ipHeader->Checksum >> 0);
	

	*buffer	+= headerLength;
	*size	-= headerLength;
	
	return 0;
}



int IpV4SetFrameHeader(uint8_t protocol, INET_IP_Type* srcIpAddr, INET_IP_Type* destIpAddr, uint16_t payloadLength, IPV4_PTR_OPTION_Type* option, IPV4_HEADER_Type* ipHeader)
{
	if (option)
	{
		if (option->Length & 0x3)	// Multiple of 4 is necessary !
			return -1;
		//if (option->Buffer == 0)	// A buffer is required
			//return -2;
		ipHeader->Option.Buffer	= option->Buffer;
		ipHeader->Option.Length	= option->Length;
	}
	else
	{
		ipHeader->Option.Buffer	= 0;
		ipHeader->Option.Length	= 0;
	}

	uint8_t ttl = IPV4_DEFTTL;
	if (protocol == IGMP_IPTYPE) { ttl = 1; }

	uint16_t headerLength = IPV4_MIN_HEADER_SIZE + ipHeader->Option.Length;
		
	ipHeader->Version				= IPV4_VERSION;
	ipHeader->HeaderLengthX32B		= headerLength >> 2;
	ipHeader->TOS.BYTE[0]			= 0;
	ipHeader->TotalLength			= headerLength + payloadLength;
	ipHeader->Identification		= GetNewIpHeaderIdentificationValue();
	ipHeader->FRAG.hw				= IPV4_HEADER_FRAG_DONOT;
	ipHeader->TimeToLive			= ttl;
	ipHeader->Protocol				= protocol;
	ipHeader->Checksum				= 0;
	ipHeader->SrcIpAddr				= *srcIpAddr;
	ipHeader->DestIpAddr			= *destIpAddr;
	
	ipHeader->PayloadLength			= payloadLength;
	
	return 0;
}



uint32_t IpV4CalcChecksumPseudoHeader(IPV4_HEADER_Type* ipHeader, uint16_t protoLength)
{
	uint32_t checksum = 0;

	checksum += (ipHeader->SrcIpAddr.BYTE[0] << 8);
	checksum += (ipHeader->SrcIpAddr.BYTE[1] << 0);
	checksum += (ipHeader->SrcIpAddr.BYTE[2] << 8);
	checksum += (ipHeader->SrcIpAddr.BYTE[3] << 0);

	checksum += (ipHeader->DestIpAddr.BYTE[0] << 8);
	checksum += (ipHeader->DestIpAddr.BYTE[1] << 0);
	checksum += (ipHeader->DestIpAddr.BYTE[2] << 8);
	checksum += (ipHeader->DestIpAddr.BYTE[3] << 0);
	
	//checksum += (0 << 8);
	checksum += (ipHeader->Protocol << 0);
	
	checksum += protoLength;
	
	return checksum;
}



_Bool IpV4IpEqual(INET_IP_Type* ip1, INET_IP_Type* ip2)
{
	uint8_t* ip1p = ip1->BYTE;
	uint8_t* ip2p = ip2->BYTE;
	
	uint32_t ip1r =   (ip1p[0] <<  0)
					| (ip1p[1] <<  8)
					| (ip1p[2] << 16)
					| (ip1p[3] << 24);

	uint32_t ip2r =   (ip2p[0] <<  0)
					| (ip2p[1] <<  8)
					| (ip2p[2] << 16)
					| (ip2p[3] << 24);
					
	return (ip1r == ip2r);
}



_Bool IpV4IpInSameSubnet(INET_IP_Type* ip1, INET_IP_Type* ip2, INET_IP_Type* snm)
{
	uint8_t* ip1p = ip1->BYTE;
	uint8_t* ip2p = ip2->BYTE;
	uint8_t* snmp = snm->BYTE;

	uint32_t ip1r =   (ip1p[0] <<  0)
					| (ip1p[1] <<  8)
					| (ip1p[2] << 16)
					| (ip1p[3] << 24);

	uint32_t ip2r =   (ip2p[0] <<  0)
					| (ip2p[1] <<  8)
					| (ip2p[2] << 16)
					| (ip2p[3] << 24);

	uint32_t snmr =   (snmp[0] <<  0)
					| (snmp[1] <<  8)
					| (snmp[2] << 16)
					| (snmp[3] << 24);

	return ((ip1r & snmr) == (ip2r & snmr));
}



_Bool IpV4IsAny(INET_IP_Type* ip)
{
	uint8_t* ipp = ip->BYTE;
		
	uint32_t ipr =    (ipp[0] <<  0)
					| (ipp[1] <<  8)
					| (ipp[2] << 16)
					| (ipp[3] << 24);

	return (ipr == 0);					
}