/*
 * UDP.c
 *
 * Created: 30.12.2018 18:03:11
 *  Author: Thomas
 */ 


#include <stdbool.h>
#include <string.h>

#include "Internet/Internet.h"
#include "UDP.h"



static uint16_t UdpCalcChecksum(IPV4_HEADER_Type* ipHeader, uint8_t* buffer, uint16_t length)
{
	uint32_t checksum = IpV4CalcChecksumPseudoHeader(ipHeader, length);
	
	checksum += (buffer[0] << 8) | buffer[1];	// ScrPort
	checksum += (buffer[2] << 8) | buffer[3];	// DestPort
	checksum += (buffer[4] << 8) | buffer[5];	// TotalLength
	//checksum += (buffer[6] << 8) | buffer[7];	// Checksum (in this case it must be zero)
	
	buffer += UDP_HEADER_SIZE;	
	length -= UDP_HEADER_SIZE;

	for (int i = 0; i < (length >> 1); i++)
	{
		checksum += (buffer[(i << 1) + 0] << 8);
		checksum += (buffer[(i << 1) + 1] << 0);
	}
	
	if (length & 0x1)
		checksum += (buffer[length & 0xfffe] << 8);
	
	while (checksum & 0xffff0000)
		checksum = (checksum >> 16) + (checksum & 0xffff);
	
	if(checksum == 0xffff)
		return checksum;

	return ~checksum;
}




int UdpReadFrame(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader, UDP_FRAME_Type* udpFrame)
{
	if (ipHeader->Protocol != UDP_IPTYPE)
		return -1;
	if (*size < ipHeader->PayloadLength)
		return -2;

	uint8_t* buf = *buffer;
	
	udpFrame->UdpHeader.ScrPort		= (buf[0] << 8) | buf[1];
	udpFrame->UdpHeader.DestPort	= (buf[2] << 8) | buf[3];
	udpFrame->UdpHeader.TotalLength	= (buf[4] << 8) | buf[5];
	udpFrame->UdpHeader.Checksum	= (buf[6] << 8) | buf[7];

	if (udpFrame->UdpHeader.Checksum)
	{
		uint16_t checksum = UdpCalcChecksum(ipHeader, buf, udpFrame->UdpHeader.TotalLength);
		if (checksum != udpFrame->UdpHeader.Checksum)
			return -3;
	}
	
	if (udpFrame->UdpHeader.TotalLength > UDP_HEADER_SIZE)
	{
		udpFrame->Data.Length = udpFrame->UdpHeader.TotalLength - UDP_HEADER_SIZE;
		udpFrame->Data.Buffer = buf + UDP_HEADER_SIZE;
	}
	else
	{
		udpFrame->Data.Length = 0;
		udpFrame->Data.Buffer = 0;
	}
	
	*buffer	+= udpFrame->UdpHeader.TotalLength;
	*size	-= udpFrame->UdpHeader.TotalLength;
	
	return 0;
}



int UdpWriteFrame(uint8_t* buffer[], uint16_t* size, IPV4_HEADER_Type* ipHeader, UDP_FRAME_Type* udpFrame)
{
	if (*size < udpFrame->UdpHeader.TotalLength)
		return -1;
	
	uint8_t* buf = *buffer;

	buf[0] = (uint8_t)(udpFrame->UdpHeader.ScrPort >> 8);
	buf[1] = (uint8_t)(udpFrame->UdpHeader.ScrPort >> 0);
	buf[2] = (uint8_t)(udpFrame->UdpHeader.DestPort >> 8);
	buf[3] = (uint8_t)(udpFrame->UdpHeader.DestPort >> 0);
	buf[4] = (uint8_t)(udpFrame->UdpHeader.TotalLength >> 8);
	buf[5] = (uint8_t)(udpFrame->UdpHeader.TotalLength >> 0);
	
	if (udpFrame->Data.Length)
		memcpy(&buf[UDP_HEADER_SIZE], udpFrame->Data.Buffer, udpFrame->Data.Length);
	
	
	udpFrame->UdpHeader.Checksum = UdpCalcChecksum(ipHeader, buf, udpFrame->UdpHeader.TotalLength);
	
	buf[6] = (uint8_t)(udpFrame->UdpHeader.Checksum >> 8);
	buf[7] = (uint8_t)(udpFrame->UdpHeader.Checksum >> 0);

	
	*buffer	+= udpFrame->UdpHeader.TotalLength;
	*size	-= udpFrame->UdpHeader.TotalLength;
	
	return 0;
} 



int UdpSetFrame(uint16_t scrPort, uint16_t destPort, UDP_PTR_DATA_Type* data, UDP_FRAME_Type* udpFrame)
{
	udpFrame->UdpHeader.ScrPort = scrPort;
	udpFrame->UdpHeader.DestPort = destPort;
	udpFrame->UdpHeader.Checksum = 0;
	
	uint16_t totalLength = UDP_HEADER_SIZE;

	if (data)
	{
		udpFrame->Data = *data;
		totalLength += udpFrame->Data.Length;
	}
	
	udpFrame->UdpHeader.TotalLength = totalLength;
	
	return 0;
}

	


