/*
 * Router.c
 *
 * Created: 8.1.2019 18:25:15
 *  Author: Thomas
 */ 


#include <stdbool.h>
#include <string.h>

#include "sam.h"

#include "misc.h"
#include "IO.h"
#include "Time/Time.h"

#include "Internet/Internet.h"
#include "Internet/EthernetII/EthernetII.h"
#include "Internet/ARP/ARP.h"
#include "Internet/IP/IPV4.h"
#include "Internet/ICMP/ICMP.h"
#include "Internet/IGMP/IGMP.h"
#include "Internet/UDP/UDP.h"
#include "Socket/Socket.h"
#include "SocketMain/SocketMain.h"
#include "StatusLed/StatusLed.h"

#include "Router.h"



#define CHECK_RETV_RET(x)	do { int ret_value_func = (x); if (ret_value_func < 0) { return; } } while(0);
#define CHECK_RETV_RETV(x)	do { int ret_value_func = (x); if (ret_value_func < 0) { return (ret_value_func); } } while(0);




typedef struct {
	_Bool				IsValid;
	_Bool				IsServer;
	_Bool				CheckIp;
	_Bool				IpIsInUse;
	ETH_VLAN_TAG_Type	vLanTag;
	INET_IP_Type		Ip;
	struct {
		_Bool			SendReport;
		int				Delay;
		int				Version;
	} IGMP;
	struct {
		int				DelayRequest;
		int				CounterRequests;
	} CHECKIP;
} ROUTER_MC_Type;



static ROUTER_MC_Type	RouterMcData[4];
static const int		NumberVlans	= 4;

static uint32_t			CyclicTasksIntervalTimer100ms;



int RouterProcessIpStartUp(Socket_Type* socket, uint8_t txBuffer[], uint16_t txBufferSize, ROUTER_MC_Type* routerMcData);

int RouterProcessIgmpQuery(Socket_Type* socket, uint8_t txBuffer[], uint16_t txBufferSize, ROUTER_MC_Type* routerMcData);

_Bool RouterFilterUpnpMessage(char message[], _Bool isServer);

// ----------------------------------------------------------------------------------------------


int RouterInit(void)
{
	
	CHECK_RETV_RETV(EthSetVLanTag(VLAN1_ROUTER_ID, 0, 0, &RouterMcData[0].vLanTag));
	RouterMcData[0].Ip			= VLAN1_ROUTER_IP;
	RouterMcData[0].IsServer	= IpV4IpInSameSubnet(&VLAN1_ROUTER_IP, &MC_SERVER_IP, &MC_SERVER_SNM);
	RouterMcData[0].IsValid		= !IpV4IsAny(&VLAN1_ROUTER_IP);
		
	CHECK_RETV_RETV(EthSetVLanTag(VLAN2_ROUTER_ID, 0, 0, &RouterMcData[1].vLanTag));
	RouterMcData[1].Ip			= VLAN2_ROUTER_IP;
	RouterMcData[1].IsServer	= IpV4IpInSameSubnet(&VLAN2_ROUTER_IP, &MC_SERVER_IP, &MC_SERVER_SNM);
	RouterMcData[1].IsValid		= !IpV4IsAny(&VLAN2_ROUTER_IP);

	CHECK_RETV_RETV(EthSetVLanTag(VLAN3_ROUTER_ID, 0, 0, &RouterMcData[2].vLanTag));
	RouterMcData[2].Ip			= VLAN3_ROUTER_IP;
	RouterMcData[2].IsServer	= IpV4IpInSameSubnet(&VLAN3_ROUTER_IP, &MC_SERVER_IP, &MC_SERVER_SNM);
	RouterMcData[2].IsValid		= !IpV4IsAny(&VLAN3_ROUTER_IP);

	CHECK_RETV_RETV(EthSetVLanTag(VLAN4_ROUTER_ID, 0, 0, &RouterMcData[3].vLanTag));
	RouterMcData[3].Ip			= VLAN4_ROUTER_IP;
	RouterMcData[3].IsServer	= IpV4IpInSameSubnet(&VLAN4_ROUTER_IP, &MC_SERVER_IP, &MC_SERVER_SNM);
	RouterMcData[3].IsValid		= !IpV4IsAny(&VLAN4_ROUTER_IP);
	
	RouterSetIpStartUp();
	
	return 0;
}


// ----------------------------------------------------------------------------------------------


int RouterSetIpStartUp(void)
{
	IntervalTimer_ms(&CyclicTasksIntervalTimer100ms, 0);
		
	for (int i = 0; i < NumberVlans; i++)
	{
		if (!RouterMcData[i].IsValid)
			continue;

		RouterMcData[i].CheckIp					= true;
		RouterMcData[i].IpIsInUse				= false;
		RouterMcData[i].CHECKIP.DelayRequest	= 20 + 1 + i;
		RouterMcData[i].CHECKIP.CounterRequests	= 0;
	}
	
	return 0;	
}


// ----------------------------------------------------------------------------------------------


int RouterProcessCyclicTasks(Socket_Type* socket, uint8_t txBuffer[], uint16_t txBufferSize)
{
	int retVal = 0;
	
	if (IntervalTimer_ms(&CyclicTasksIntervalTimer100ms, 100) == 0)
		return 0;
	
	
	WNET_CRB_PHYCFGR_Type phyCfg;
	
	CHECK_RETV_RETV(WNetReadByte(WNET_CRB_PHYCFGR_ADDRESS, WNET_FRAME_CONTROL_PHASE_BSEL_CREG, &phyCfg.by));
	
	if (phyCfg.b.LNK == WNET_CRB_PHYCFGR_LNK_DOWN)
	{
		RouterSetIpStartUp();
		return 0;
	}
			
	
	for (int i = 0; i < NumberVlans; i++)
	{
		if (!RouterMcData[i].IsValid)
			continue;
			
		if (RouterMcData[i].IpIsInUse)
		{
			StatusLedSignalError();	
			continue;
		}
	
		if (RouterMcData[i].CheckIp)
		{
			if(RouterProcessIpStartUp(socket, txBuffer, txBufferSize, &RouterMcData[i]) < 0)
				{ retVal = -1; continue; }
		}
		else
		{
			if (RouterProcessIgmpQuery(socket, txBuffer, txBufferSize, &RouterMcData[i]) < 0)
				{ retVal = -1; continue; }
		}	
	}
	
	return retVal;
}


// ----------------------------------------------------------------------------------------------


int RouterProcessIpStartUp(Socket_Type* socket, uint8_t txBuffer[], uint16_t txBufferSize, ROUTER_MC_Type* routerMcData)
{
	routerMcData->CHECKIP.DelayRequest += -1;
	if (routerMcData->CHECKIP.DelayRequest > 0)
		return 0;

	routerMcData->CHECKIP.DelayRequest = 10;

	
	if (routerMcData->CHECKIP.CounterRequests >= 3)
	{
		routerMcData->CheckIp = false;
			
		CHECK_RETV_RETV(SendGratuitousArp(socket, txBuffer, txBufferSize, &routerMcData->vLanTag, &routerMcData->Ip));
	
		routerMcData->IGMP.SendReport	= true;
		routerMcData->IGMP.Version		= 2;
		routerMcData->IGMP.Delay		= 10;
		
		return 0;
	}
	routerMcData->CHECKIP.CounterRequests += 1;
		
	CHECK_RETV_RETV(SendDummyArp(socket, txBuffer, txBufferSize, &routerMcData->vLanTag, &routerMcData->Ip));
	
	return 0;
}


// ----------------------------------------------------------------------------------------------


int RouterProcessIgmpQuery(Socket_Type* socket, uint8_t txBuffer[], uint16_t txBufferSize, ROUTER_MC_Type* routerMcData)
{
	if (!routerMcData->IGMP.SendReport)
		return 0;
		
	routerMcData->IGMP.Delay += -1;
	if (routerMcData->IGMP.Delay > 0)
		return 0;
		
	routerMcData->IGMP.SendReport = false;
		
	uint8_t* pTxBuf		= txBuffer;
	uint16_t sizeTxBuf	= txBufferSize;

	IGMP_FRAME_Type igmpFrameTx;
	if (routerMcData->IGMP.Version == 1)
	{
		CHECK_RETV_RETV(IgmpSetReportFrameV1(&MC_GROUP, &igmpFrameTx));
	}
	else
	{
		CHECK_RETV_RETV(IgmpSetReportFrameV2(&MC_GROUP, &igmpFrameTx));
	}
	
	IPV4_HEADER_Type ipHeaderTx;
	uint16_t ipPayload = igmpFrameTx.Length;

	CHECK_RETV_RETV(IpV4SetFrameHeader( IGMP_IPTYPE,
										&routerMcData->Ip,
										&MC_GROUP,
										ipPayload,
										IpV4GetRouterAlertOptionPtr(),
										&ipHeaderTx ));

	INET_MAC_Type McGroupMac;
	CHECK_RETV_RETV(EthMacFromMulticastGroupIpV4(&McGroupMac, &MC_GROUP));

	ETH_L2_FRAMEH_Type ethFrameTx;
	CHECK_RETV_RETV(EthSetFrameHeader(  &McGroupMac,
										&MY_MAC,
										IPV4_ETHERTYPE,
										&routerMcData->vLanTag,
										&ethFrameTx));

		
	CHECK_RETV_RETV(EthWriteFrameHeader(&pTxBuf, &sizeTxBuf, &ethFrameTx));
		
	CHECK_RETV_RETV(IpV4WriteFrameHeader(&pTxBuf, &sizeTxBuf, &ipHeaderTx));

	CHECK_RETV_RETV(IgmpWriteFrame(&pTxBuf, &sizeTxBuf, &igmpFrameTx));
		
	CHECK_RETV_RETV(SocketSendMacRaw(socket, txBuffer, pTxBuf - txBuffer));
	
	return 0;
}


// ----------------------------------------------------------------------------------------------


int RouterReceiveMacFrame(Socket_Type* socket, uint8_t rxBuffer[], uint16_t rxBufferSize, uint16_t receivedBytes, uint8_t txBuffer[], uint16_t txBufferSize)
{
	uint8_t* pRxBuf		= rxBuffer;
	uint8_t* pTxBuf		= txBuffer;
	uint16_t sizeRxBuf	= receivedBytes;
	uint16_t sizeTxBuf	= txBufferSize;
	ETH_L2_FRAMEH_Type	ethFrameRx;
	ETH_L2_FRAMEH_Type	ethFrameTx;
	ROUTER_MC_Type*		pRouterMcData = 0;
	
	
	CHECK_RETV_RETV(EthReadFrameHeader(&pRxBuf, &sizeRxBuf, &ethFrameRx));
		
	
	// ##### Filter VLAN
	for (int i = 0; i < NumberVlans; i++)
	{
		if (RouterMcData[i].IsValid)
			if (ethFrameRx.VLAN.TCI.b.VID == RouterMcData[i].vLanTag.TCI.b.VID)
			{
				pRouterMcData = &RouterMcData[i];
				break;
			}
				
		if ((i + 1) == NumberVlans)
			return 0;
	}
	
	
	// If this IP is in use then prevent all further actions
	if (pRouterMcData->IpIsInUse)
		return 0;
		
	
	INET_MAC_Type		sendToMacAddr		= ethFrameRx.SrcMacAddr;
	ETH_VLAN_TAG_Type	sendToVlanTag		= ethFrameRx.VLAN;
		
	
	// ##### Filter ARP
	if (ethFrameRx.EthType == ARP_ETHERTYPE)
	{
		ARP_FRAME_Type arpFrameRx;
		CHECK_RETV_RETV(ArpReadFrame(&pRxBuf, &sizeRxBuf, &arpFrameRx));
		


		if (pRouterMcData->CheckIp)
		{
			if (arpFrameRx.Operation == ARP_OPERATION_RESP)
				if (IpV4IpEqual(&arpFrameRx.SenderIpAddr, &pRouterMcData->Ip))
					pRouterMcData->IpIsInUse = true;

			return 0;
		}

		
		
		if (arpFrameRx.Operation != ARP_OPERATION_REQ)
			return 0;

		if (!IpV4IpEqual(&arpFrameRx.TargetIpAddr, &pRouterMcData->Ip))
			return 0;	

		INET_IP_Type myIpAddr	= arpFrameRx.TargetIpAddr;


		ARP_FRAME_Type arpFrameTx;
		CHECK_RETV_RETV(ArpSetFrame( ARP_OPERATION_RESP,
									 &MY_MAC,
									 &myIpAddr,
									 &arpFrameRx.SenderMacAddr,
									 &arpFrameRx.SenderIpAddr,
									 &arpFrameTx));

		CHECK_RETV_RETV(EthSetFrameHeader( &sendToMacAddr, 
										   &MY_MAC, 
										   ARP_ETHERTYPE, 
										   &sendToVlanTag, 
										   &ethFrameTx));


		CHECK_RETV_RETV(EthWriteFrameHeader(&pTxBuf, &sizeTxBuf, &ethFrameTx));

		CHECK_RETV_RETV(ArpWriteFrame(&pTxBuf, &sizeTxBuf, &arpFrameTx));
		
		//CHECK_RETV_RETV(SocketSendMacRaw(socket, txBuffer, pTxBuf - txBuffer));
				
		CHECK_RETV_RETV(SocketBeginSendMacRaw(socket, txBuffer, pTxBuf - txBuffer, 0));
		
		return 0;
	}
	
			
	
	// ##### Filter IP
	if (ethFrameRx.EthType != IPV4_ETHERTYPE)
		return 0;
	
	if (pRouterMcData->CheckIp)
		return 0;
		
	IPV4_HEADER_Type ipHeaderRx;
	CHECK_RETV_RETV(IpV4ReadFrameHeader(&pRxBuf, &sizeRxBuf, &ipHeaderRx));
	
	uint8_t				ipProtocol		= ipHeaderRx.Protocol;
	INET_IP_Type		remoteIpAddr	= ipHeaderRx.SrcIpAddr;
	IPV4_HEADER_Type	ipHeaderTx;

				
		
	// ##### Filter ICMP
	if (ipProtocol == ICMP_IPTYPE)
	{
		if (IcmpGetType(&ipHeaderRx, pRxBuf, sizeRxBuf) != ICMP_TYPE_ECHO_REQ)
			return 0;

		if (!IpV4IpEqual(&ipHeaderRx.DestIpAddr, &pRouterMcData->Ip))
			return 0;

		ICMP_ECHO_FRAME_Type echoFrameRx;
		CHECK_RETV_RETV(IcmpReadEchoFrame(&pRxBuf, &sizeRxBuf, &ipHeaderRx, &echoFrameRx));


		ICMP_ECHO_FRAME_Type echoFrameTx;
		CHECK_RETV_RETV(IcmpSetEchoFrame( ICMP_TYPE_ECHO_REP,
										  echoFrameRx.Identifier,
										  echoFrameRx.SeqNumber,
										  &echoFrameRx.Data,
										  &echoFrameTx));
		
		uint16_t ipPayload = echoFrameTx.TotalLength;
		
		CHECK_RETV_RETV(IpV4SetFrameHeader( ICMP_IPTYPE,
										    &pRouterMcData->Ip,
										    &remoteIpAddr,
										    ipPayload,
										    0,
										    &ipHeaderTx ));

		CHECK_RETV_RETV(EthSetFrameHeader( &sendToMacAddr, 
						  				   &MY_MAC, 
										   IPV4_ETHERTYPE, 
										   &sendToVlanTag, 
										   &ethFrameTx));


		CHECK_RETV_RETV(EthWriteFrameHeader(&pTxBuf, &sizeTxBuf, &ethFrameTx));

		CHECK_RETV_RETV(IpV4WriteFrameHeader(&pTxBuf, &sizeTxBuf, &ipHeaderTx));
		
		CHECK_RETV_RETV(IcmpWriteEchoFrame(&pTxBuf, &sizeTxBuf, &echoFrameTx))
		
		//CHECK_RETV_RETV(SocketSendMacRaw(socket, txBuffer, pTxBuf - txBuffer));
		
		CHECK_RETV_RETV(SocketBeginSendMacRaw(socket, txBuffer, pTxBuf - txBuffer, 0));

		return 0;
	}
	

		
	// ##### Filter IGMP
	if (ipProtocol == IGMP_IPTYPE)
	{	
		if (!IpV4IpEqual(&ipHeaderRx.DestIpAddr, &IGMP_IP_QUERY))
			return 0;
			
		IGMP_FRAME_Type igmpFrameRx;
		int igmpVersion;
		
		igmpVersion = IgmpReadFrame(&pRxBuf, &sizeRxBuf, &ipHeaderRx, &igmpFrameRx);
		if ((igmpVersion < 1) || (igmpVersion > 2))
			return 0;
			
		if (igmpFrameRx.Type != IGMP_MEMBERSHIP_QUERY)
			return 0;
		
		pRouterMcData->IGMP.SendReport	= true;
		pRouterMcData->IGMP.Version		= igmpVersion;
		
		if (igmpVersion == 1)
			pRouterMcData->IGMP.Delay = 1;
		else
			pRouterMcData->IGMP.Delay = igmpFrameRx.Code >> 1;
		
		return 0;
	}
	

	
	// ##### Is routing allowed (TTL > 1)
	if (ipHeaderRx.TimeToLive <= 1)
		return 0;
		
		
	
	// ##### Filter Multicast
	if(!EthIsMulticastFromGroupIpV4(&ethFrameRx.DestMacAddr, &MC_GROUP))
		return 0;
	
	if (!IpV4IpEqual(&ipHeaderRx.DestIpAddr, &MC_GROUP))
		return 0;
	
	
	
	// ##### Filter UDP
	if (ipProtocol != UDP_IPTYPE)
		return 0;
		
	UDP_FRAME_Type udpFrameRx;
		
	CHECK_RETV_RETV(UdpReadFrame(&pRxBuf, &sizeRxBuf, &ipHeaderRx, &udpFrameRx));
	
	if (udpFrameRx.UdpHeader.DestPort != MC_SERVER_PORT)	
		return 0;
	
	if (udpFrameRx.Data.Length == 0)
		return 0;
	
	// Terminate UDP-Data with 0 (Make a string)
	if ((pRxBuf - rxBuffer) >= rxBufferSize)
		return 0;

	pRxBuf[0] = 0;	
		
	
	
	// ##### UPnP
		
		
	_Bool IsFromServerNet = IpV4IpInSameSubnet(&pRouterMcData->Ip, &MC_SERVER_IP, &MC_SERVER_SNM);
	
	
	// From server subnet only the server is routed
	if(IsFromServerNet && !IpV4IpEqual(&ipHeaderRx.SrcIpAddr, &MC_SERVER_IP))
		return 0;
	
	
	if(!RouterFilterUpnpMessage((char*)udpFrameRx.Data.Buffer, IsFromServerNet))
		return 0;


	// Reduce lifetime
	ipHeaderRx.TimeToLive += -1;		
		

	INET_MAC_Type McGroupMac;
	CHECK_RETV_RETV(EthMacFromMulticastGroupIpV4(&McGroupMac, &MC_GROUP));
	
	int retVal = 0;
		
	for (int i = 0; i < NumberVlans; i++)
	{
		if (!RouterMcData[i].IsValid)
			continue;

		if (RouterMcData[i].IpIsInUse)
			continue;

		if (RouterMcData[i].CheckIp)
			continue;
			
		// Is this the origin 
		if (RouterMcData[i].vLanTag.TCI.b.VID == pRouterMcData->vLanTag.TCI.b.VID)
			continue;

		// Route client only to the server
		if (!IsFromServerNet && !RouterMcData[i].IsServer)
			continue;
		
		
		pTxBuf		= txBuffer;
		sizeTxBuf	= txBufferSize;
		
		if(EthSetFrameHeader(&McGroupMac, &MY_MAC, IPV4_ETHERTYPE, &RouterMcData[i].vLanTag, &ethFrameTx) < 0)
			{ retVal = -1; continue; }
				
		if(EthWriteFrameHeader(&pTxBuf, &sizeTxBuf, &ethFrameTx) < 0)
			{ retVal = -1; continue; }
			
		if(IpV4WriteFrameHeader(&pTxBuf, &sizeTxBuf, &ipHeaderRx) < 0)
			{ retVal = -1; continue; }
		
		if(UdpWriteFrame(&pTxBuf, &sizeTxBuf, &ipHeaderRx, &udpFrameRx) < 0)
			{ retVal = -1; continue; }
	
		if(SocketSendMacRaw(socket, txBuffer, pTxBuf - txBuffer) < 0)
			{ retVal = -1; continue; }
	}
	
	return retVal;
}


// ----------------------------------------------------------------------------------------------


_Bool RouterFilterUpnpMessage(char* message, _Bool isServer)
{
	// MediaServer
	// "urn:schemas-upnp-org:device:MediaServer:1"
	// "urn:schemas-upnp-org:service:ContentDirectory:1"
	// "urn:schemas-upnp-org:service:ConnectionManager:1"
	// "urn:schemas-upnp-org:service:AVTransport:1" Optional
		
	// MediaRenderer
	// "urn:schemas-upnp-org:device:MediaRenderer:1"
	// "urn:schemas-upnp-org:service:RenderingControl:1"
	// "urn:schemas-upnp-org:service:ConnectionManager:1"
	// "urn:schemas-upnp-org:service:AVTransport:1" Optional
	
	
	static const char StrMSearch[]		= "M-SEARCH * HTTP/1.1\0";
	static const char StrNotify[]		= "NOTIFY * HTTP/1.1\0";
	static const char StrDMServer[]		= "urn:schemas-upnp-org:device:MediaServer:1\0";
	static const char StrDMRenderer[]	= "urn:schemas-upnp-org:device:MediaRenderer:1\0";
	static const char StrSCDirectory[]	= "urn:schemas-upnp-org:service:ContentDirectory:1\0";
	static const char StrSRControl[]	= "urn:schemas-upnp-org:service:RenderingControl:1\0";
	static const char StrSCManager[]	= "urn:schemas-upnp-org:service:ConnectionManager:1\0";
	static const char StrSATransport[]	= "urn:schemas-upnp-org:service:AVTransport:1\0"; // Optional

	
	// MediaServer
	if(isServer)
	{
		// MediaServer "NOTIFY"
		if (strstr(message, StrNotify))
		{
			// "device:MediaServer"
			if (strstr(message, StrDMServer))
				return true;
		
			// "service:ContentDirectory"
			if (strstr(message, StrSCDirectory))
				return true;
		
			// "service:ConnectionManager"
			if (strstr(message, StrSCManager))
				return true;
		
			// "service:AVTransport" (Optional)
			if (strstr(message, StrSATransport))
				return true;
		}
		
		return false;
	}


	// MediaRenderer "M-SEARCH"
	if (strstr(message, StrMSearch))
	{
		// "device:MediaServer"
		if (strstr(message, StrDMServer))
			return true;
		
		return false;
	}
		
	
	// MediaRenderer "NOTIFY"
	if (strstr(message, StrNotify))
	{
		// "device:MediaRenderer"
		if (strstr(message, StrDMRenderer))
			return true;
		
		// "service:RenderingControl"
		if (strstr(message, StrSRControl))
			return true;
				
		// "service:ConnectionManager"
		if (strstr(message, StrSCManager))
			return true;
		
		// "service:AVTransport" (Optional)
		if (strstr(message, StrSATransport))
			return true;

		return false;
	}

	
	return false;	
}


// ----------------------------------------------------------------------------------------------



	