/*
 * Socket.c
 *
 * Created: 17.11.2018 14:55:08
 *  Author: Thomas
 */ 

#include <stdbool.h>
#include "sam.h"

#include "misc.h"
#include "IO.h"
#include "Time/Time.h"
#include "Socket.h"
#include "StatusLed/StatusLed.h"


#define CHECK_RETV_RETV(x)	do { int ret_value_func = (x); if (ret_value_func < 0) { return (ret_value_func); } } while(0);
#define CHECK_RETV_RET(x)	do { int ret_value_func = (x); if (ret_value_func < 0) { return; } } while(0);


static _Bool ChannelInUse[8];

static FLASH WNET_MAC_Type BroadcastMac = {{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }};


static Timer_Type soTimeout;

static void SocketSetTimeout(int ms)
{
	TimerStart_ms(ms, &soTimeout);
}

static int SocketCheckTimeout(void)
{
	if ((int32_t)soTimeout.TimeTicks < 0) return SOCKET_NOERR;
	
	if(TimerTest(&soTimeout)) return SOCKET_ERROR_TIMEOUT;

	return SOCKET_NOERR;	
}


// ----------------------------------------------------------------------------------------------


int SocketOpen(Socket_Type* socket, uint8_t channel, SocketType_Type socketType, SocketIpProtocol_Type socketIpProtocol, uint16_t port, int timeOut_ms)
{
	if (channel < 0 || channel > 7)
		return SOCKET_ERROR_PARAMETER2;
	else if (ChannelInUse[channel])
		return SOCKET_ERROR_CHANNEL_IN_USE;
	else if ((socketType == SocketTypeIp) && (socketIpProtocol == SocketIpProtocolNothing))
		return SOCKET_ERROR_PARAMETER5;

	
	SocketSetTimeout(timeOut_ms);
	
	SocketStatus_Type socketStatus;
	uint8_t sxreg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(channel);

	
	switch (socketType)
	{
		case SocketTypeUdp:
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_MR_ADDRESS, sxreg, WNET_SRB_MR_P_UDP));
			CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_SPORT_ADDRESS, sxreg, port));
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, sxreg, WNET_SRB_SCR_CMD_OPEN));
			do{
				CHECK_RETV_RETV(WNetReadByte(WNET_SRB_SSR_ADDRESS, sxreg, (uint8_t*)&socketStatus));
				CHECK_RETV_RETV(SocketCheckTimeout());
			} while(socketStatus != WNET_SRB_SSR_STAT_SOCK_UDP);
		break;
		
		
		case SocketTypeIp:
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_MR_ADDRESS, sxreg, WNET_SRB_MR_P_IPRAW));
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SIPPR_ADDRESS, sxreg, socketIpProtocol));
			CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_SPORT_ADDRESS, sxreg, port));
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, sxreg, WNET_SRB_SCR_CMD_OPEN));
			do{
				CHECK_RETV_RETV(WNetReadByte(WNET_SRB_SSR_ADDRESS, sxreg, (uint8_t*)&socketStatus));
				CHECK_RETV_RETV(SocketCheckTimeout());
			} while(socketStatus != WNET_SRB_SSR_STAT_SOCK_IPRAW);
		break;


		case SocketTypeMacRaw:
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_MR_ADDRESS, sxreg, WNET_SRB_MR_P_MACRAW | WNET_SRB_MR_MACRAW_MFEN | WNET_SRB_MR_MACRAW_MIP6B));
			CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, sxreg, WNET_SRB_SCR_CMD_OPEN));
			do{
				CHECK_RETV_RETV(WNetReadByte(WNET_SRB_SSR_ADDRESS, sxreg, (uint8_t*)&socketStatus));
				CHECK_RETV_RETV(SocketCheckTimeout());
			} while(socketStatus != WNET_SRB_SSR_STAT_SOCK_MACRAW);
		break;
		
		
		default:
			return SOCKET_ERROR_PARAMETER1;
		break;
	}
	
	

	socket->Channel		= channel;
	socket->Type		= socketType;
	socket->IpProtocol	= socketIpProtocol;
	socket->Status		= socketStatus;
	
	ChannelInUse[channel] = true;
	
	return SOCKET_NOERR;
}


int SocketClose(Socket_Type* socket)
{
	ChannelInUse[socket->Channel] = false;
	
	socket->Channel		= -1;
	socket->Type		= SocketTypeNothing;
	socket->IpProtocol	= SocketIpProtocolNothing;
	socket->Status		= SocketStatusNothing;
	
	return SOCKET_NOERR;
}


// ----------------------------------------------------------------------------------------------


int GetSxReceivedSize(uint8_t controlPhaseBsel, uint16_t* rxSize)
{
	uint16_t rxSizeCmp;
	
	do {
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RSR_ADDRESS, controlPhaseBsel, rxSize));
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RSR_ADDRESS, controlPhaseBsel, &rxSizeCmp));
	} while (*rxSize != rxSizeCmp);
	
	return SOCKET_NOERR;
}



int GetSxTxBufferFreeSize(uint8_t controlPhaseBsel, uint16_t* txFreeSize)
{
	uint16_t txFreeSizeCmp;
	
	do {
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_FSR_ADDRESS, controlPhaseBsel, txFreeSize));
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_FSR_ADDRESS, controlPhaseBsel, &txFreeSizeCmp));
	} while (*txFreeSize != txFreeSizeCmp);
	
	return SOCKET_NOERR;
}


// Wait until the last send is finished other vise the "TX_RD" will updated with the new "TX_WR"
// This cause that none bytes will be transmit. (Single buffered)
int WaitUntilSxLastSendFinished(uint8_t controlPhaseBsel)
{
	uint16_t txWr, txRd;

	do {
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_WR_ADDRESS, controlPhaseBsel, &txWr));
		CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_RD_ADDRESS, controlPhaseBsel, &txRd));
	} while (txWr != txRd);
	
	return SOCKET_NOERR;
}


// ----------------------------------------------------------------------------------------------


static SocketUdpHeader_Type SxUdpHeader;

SocketUdpHeader_Type SocketGetReceiveUdpHeader(void)
{
	return SxUdpHeader;
}


static struct {
	_Bool IsActive;
	uint16_t RxRd;
	uint16_t TxWr;
	uint8_t ScrCmd;
	SocketAsyncCallBack_Type CallBack;
	SOCKET_RW_ASYNC_RESULT_Type AsyncResult;
} SxAsyncInfo;

	
int SocketGetAsyncResult(SOCKET_RW_ASYNC_RESULT_Type* asyncResult)
{
	if (asyncResult == 0)
		return SOCKET_ERROR_PARAMETER1;
	
	uint32_t cs;
	WNetEnterCriticalSectionAsyncResult(&cs);
	
	*asyncResult = SxAsyncInfo.AsyncResult;
	
	if (!SxAsyncInfo.AsyncResult.IsBusy)
	{
		SxAsyncInfo.AsyncResult.IsCompleted		= false;
		SxAsyncInfo.AsyncResult.AsyncState		= 0;
		SxAsyncInfo.AsyncResult.CallBackState	= 0;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	}
	
	WNetLeaveCriticalSectionAsyncResult(&cs);

	asyncResult->IsReady = (!asyncResult->IsBusy) && (!asyncResult->IsCompleted);

	return SOCKET_NOERR;
}



static void AsyncCallBackRead(void)
{
	SxAsyncInfo.IsActive = false;
	
	SxAsyncInfo.AsyncResult.IsBusy		= false;
	SxAsyncInfo.AsyncResult.IsCompleted	= true;
	SxAsyncInfo.AsyncResult.AsyncState	= WNetGetAsyncResult(false).AsyncState;
	
	if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
	{
		uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(SxAsyncInfo.AsyncResult.Socket->Channel);
	
		if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
		{
			SxAsyncInfo.AsyncResult.AsyncState = WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, SxAsyncInfo.RxRd);
		}
		
		if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
		{
			SxAsyncInfo.AsyncResult.AsyncState = WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, SxAsyncInfo.ScrCmd);
		}
	}

	if (SxAsyncInfo.CallBack)
	{
		SxAsyncInfo.AsyncResult.CallBackState	= SxAsyncInfo.CallBack(&SxAsyncInfo.AsyncResult);
		SxAsyncInfo.AsyncResult.IsCompleted		= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	}

	return; 
}


static void AsyncCallBackWrite(void)
{
	SxAsyncInfo.IsActive = false;
	
	SxAsyncInfo.AsyncResult.IsBusy		= false;
	SxAsyncInfo.AsyncResult.IsCompleted	= true;
	SxAsyncInfo.AsyncResult.AsyncState	= WNetGetAsyncResult(false).AsyncState;
	
	if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
	{
		uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(SxAsyncInfo.AsyncResult.Socket->Channel);
		
		if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
		{
			SxAsyncInfo.AsyncResult.AsyncState = WaitUntilSxLastSendFinished(SxReg);
		}
		
		if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
		{
			SxAsyncInfo.AsyncResult.AsyncState = WNetWriteHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, SxAsyncInfo.TxWr);
		}
		
		if(SxAsyncInfo.AsyncResult.AsyncState == WNET_NOERR)
		{
			SxAsyncInfo.AsyncResult.AsyncState = WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, SxAsyncInfo.ScrCmd);
		}
	}

	if (SxAsyncInfo.CallBack)
	{
		SxAsyncInfo.AsyncResult.CallBackState	= SxAsyncInfo.CallBack(&SxAsyncInfo.AsyncResult);
		SxAsyncInfo.AsyncResult.IsCompleted		= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	}

	return;
}


// ----------------------------------------------------------------------------------------------


int SocketBeginReceiveUdp(Socket_Type* socket, uint8_t* buffer, uint16_t bufferSize, SocketAsyncCallBack_Type callBack)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;
	
	SxAsyncInfo.AsyncResult.IsCompleted		= false;
	SxAsyncInfo.AsyncResult.IsBusy			= false;
	SxAsyncInfo.AsyncResult.AsyncState		= WNET_NOERR;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	SxAsyncInfo.AsyncResult.Socket = 0;
	
	SxUdpHeader.SOURCE.Ip		= IP_ANY;
	SxUdpHeader.SOURCE.Port		= 0;
	SxUdpHeader.DataLength		= 0;
	
	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxRxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXRXB(socket->Channel);

	uint16_t rxSize;
	CHECK_RETV_RETV(GetSxReceivedSize(SxReg, &rxSize));
	if (rxSize == 0) { return SOCKET_NOERR; }

	uint16_t rxRd;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, &rxRd));
	
	uint8_t udpHeaderBuffer[8];
	CHECK_RETV_RETV(WNetReadBytes(rxRd, SxRxB, udpHeaderBuffer, sizeof(udpHeaderBuffer)));
	rxRd += sizeof(udpHeaderBuffer);
	
	SxUdpHeader.SOURCE.Ip.BYTE[0]	=  udpHeaderBuffer[0];
	SxUdpHeader.SOURCE.Ip.BYTE[1]	=  udpHeaderBuffer[1];
	SxUdpHeader.SOURCE.Ip.BYTE[2]	=  udpHeaderBuffer[2];
	SxUdpHeader.SOURCE.Ip.BYTE[3]	=  udpHeaderBuffer[3];
	SxUdpHeader.SOURCE.Port			= (udpHeaderBuffer[4] << 8) | udpHeaderBuffer[5];
	SxUdpHeader.DataLength			= (udpHeaderBuffer[6] << 8) | udpHeaderBuffer[7];

	if(SxUdpHeader.DataLength > bufferSize)
	{
		CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + SxUdpHeader.DataLength));

		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));

		return SOCKET_ERROR_RBUFFER_SIZE;
	}	

	SxAsyncInfo.RxRd		= rxRd + SxUdpHeader.DataLength;
	SxAsyncInfo.ScrCmd		= WNET_SRB_SCR_CMD_RECV;
	SxAsyncInfo.CallBack	= callBack;
	SxAsyncInfo.IsActive	= true;

	SxAsyncInfo.AsyncResult.IsBusy			= true;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= SxUdpHeader.DataLength;
	SxAsyncInfo.AsyncResult.Socket			= socket;

	int ret = WNetBeginReadBytes(rxRd, SxRxB, buffer, SxUdpHeader.DataLength, AsyncCallBackRead);
	if (ret)
	{
		SxAsyncInfo.IsActive	= false;
		
		SxAsyncInfo.AsyncResult.IsBusy			= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
		
		return ret;
	}
	
	return SOCKET_NOERR_ASYNCH_RUNNING;
}



int SocketReceiveUdp(Socket_Type* socket, uint8_t* buffer, uint16_t bufferSize, uint16_t* receivedBytes)
{
	*receivedBytes = 0;
	
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;

	SxUdpHeader.SOURCE.Ip	= IP_ANY;
	SxUdpHeader.SOURCE.Port	= 0;
	SxUdpHeader.DataLength	= 0;
	
	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxRxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXRXB(socket->Channel);

	uint16_t rxSize;
	CHECK_RETV_RETV(GetSxReceivedSize(SxReg, &rxSize));
	if (rxSize == 0) { return SOCKET_NOERR; }

	uint16_t rxRd;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, &rxRd));
	
	uint8_t udpHeaderBuffer[8];
	CHECK_RETV_RETV(WNetReadBytes(rxRd, SxRxB, udpHeaderBuffer, sizeof(udpHeaderBuffer)));
	rxRd += sizeof(udpHeaderBuffer);
	
	SxUdpHeader.SOURCE.Ip.BYTE[0]	=  udpHeaderBuffer[0];
	SxUdpHeader.SOURCE.Ip.BYTE[1]	=  udpHeaderBuffer[1];
	SxUdpHeader.SOURCE.Ip.BYTE[2]	=  udpHeaderBuffer[2];
	SxUdpHeader.SOURCE.Ip.BYTE[3]	=  udpHeaderBuffer[3];
	SxUdpHeader.SOURCE.Port			= (udpHeaderBuffer[4] << 8) | udpHeaderBuffer[5];
	SxUdpHeader.DataLength			= (udpHeaderBuffer[6] << 8) | udpHeaderBuffer[7];

	if (SxUdpHeader.DataLength > bufferSize)
	{
		CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + SxUdpHeader.DataLength));

		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));

		return SOCKET_ERROR_RBUFFER_SIZE;
	}
	
	CHECK_RETV_RETV(WNetReadBytes(rxRd, SxRxB, buffer, SxUdpHeader.DataLength));
	
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + SxUdpHeader.DataLength));

	CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));
	
	*receivedBytes = SxUdpHeader.DataLength;
	
	return SOCKET_NOERR;
}


// ----------------------------------------------------------------------------------------------


int SocketBeginSendTo(Socket_Type* socket, uint8_t* buffer, uint16_t count, SocketEndPoint_Type* endPoint, SocketAsyncCallBack_Type callBack)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;

	SxAsyncInfo.AsyncResult.IsCompleted		= false;
	SxAsyncInfo.AsyncResult.IsBusy			= false;
	SxAsyncInfo.AsyncResult.AsyncState		= WNET_NOERR;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	SxAsyncInfo.AsyncResult.Socket = 0;

	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxTxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXTXB(socket->Channel);

	uint16_t txFreeSize;
	CHECK_RETV_RETV(GetSxTxBufferFreeSize(SxReg, &txFreeSize));
	if (txFreeSize < count) { return SOCKET_ERROR_TBUFFER_SIZE; }
	
	CHECK_RETV_RETV(WNetWriteIpType(WNET_SRB_DIPR_ADDRESS, SxReg, &endPoint->Ip));
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_DPORT_ADDRESS, SxReg, endPoint->Port));

	// Broadcast ?
	_Bool broadcast = (endPoint->Ip.IP[3] == 255);
	if (!broadcast)
	{
		SxAsyncInfo.ScrCmd = WNET_SRB_SCR_CMD_SEND;
	}
	else
	{
		CHECK_RETV_RETV(WNetWriteMacType(WNET_SRB_DHAR_ADDRESS, SxReg, &BroadcastMac));
		SxAsyncInfo.ScrCmd = WNET_SRB_SCR_CMD_SEND_MAC;
	}

	uint16_t txWr;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, &txWr));
	
	SxAsyncInfo.TxWr		= txWr + count;
	SxAsyncInfo.CallBack	= callBack;
	SxAsyncInfo.IsActive	= true;

	SxAsyncInfo.AsyncResult.IsBusy			= true;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= count;
	SxAsyncInfo.AsyncResult.Socket			= socket;

	int ret = WNetBeginWriteBytes(txWr, SxTxB, buffer, count, AsyncCallBackWrite);
	if (ret)
	{
		SxAsyncInfo.IsActive	= false;
		
		SxAsyncInfo.AsyncResult.IsBusy			= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
		
		return ret;
	}

	return SOCKET_NOERR_ASYNCH_RUNNING;
}



int SocketSendTo(Socket_Type* socket, uint8_t* buffer, uint16_t count, SocketEndPoint_Type* endPoint)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;

	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxTxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXTXB(socket->Channel);	

	uint16_t txFreeSize;
	CHECK_RETV_RETV(GetSxTxBufferFreeSize(SxReg, &txFreeSize));
	if (txFreeSize < count) { return SOCKET_ERROR_TBUFFER_SIZE; }
	
	CHECK_RETV_RETV(WNetWriteIpType(WNET_SRB_DIPR_ADDRESS, SxReg, &endPoint->Ip));
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_DPORT_ADDRESS, SxReg, endPoint->Port));

	// Broadcast ?
	_Bool broadcast = (endPoint->Ip.IP[3] == 255);
	if (broadcast)
	{
		CHECK_RETV_RETV(WNetWriteMacType(WNET_SRB_DHAR_ADDRESS, SxReg, &BroadcastMac));
	}

	uint16_t txWr;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, &txWr));
	CHECK_RETV_RETV(WNetWriteBytes(txWr, SxTxB, buffer, count));

	CHECK_RETV_RETV(WaitUntilSxLastSendFinished(SxReg));
	
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, txWr + count));

	if (!broadcast)
	{
		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_SEND));
	}
	else
	{
		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_SEND_MAC));
	}

	return SOCKET_NOERR;
}


// ----------------------------------------------------------------------------------------------


int SocketBeginReceiveMacRaw(Socket_Type* socket, uint8_t* buffer, uint16_t bufferSize, SocketAsyncCallBack_Type callBack)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;
	
	SxAsyncInfo.AsyncResult.IsCompleted		= false;
	SxAsyncInfo.AsyncResult.IsBusy			= false;
	SxAsyncInfo.AsyncResult.AsyncState		= WNET_NOERR;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	SxAsyncInfo.AsyncResult.Socket = 0;
	
	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxRxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXRXB(socket->Channel);

	uint16_t rxSize;
	CHECK_RETV_RETV(GetSxReceivedSize(SxReg, &rxSize));
	if (rxSize == 0) { return SOCKET_NOERR; }

	uint16_t rxRd;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, &rxRd));
	
	uint16_t dataLength;
	CHECK_RETV_RETV(WNetReadHWord(rxRd, SxRxB, &dataLength));
	dataLength	+= -2;
	rxRd		+=  2;

	if (dataLength > bufferSize)
	{
		CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + dataLength));

		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));

		return SOCKET_ERROR_RBUFFER_SIZE;
	}

	SxAsyncInfo.RxRd		= rxRd + dataLength;
	SxAsyncInfo.ScrCmd		= WNET_SRB_SCR_CMD_RECV;
	SxAsyncInfo.CallBack	= callBack;
	SxAsyncInfo.IsActive	= true;

	SxAsyncInfo.AsyncResult.IsBusy			= true;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= dataLength;
	SxAsyncInfo.AsyncResult.Socket			= socket;

	int ret = WNetBeginReadBytes(rxRd, SxRxB, buffer, dataLength, AsyncCallBackRead);
	if (ret)
	{
		SxAsyncInfo.IsActive	= false;
		
		SxAsyncInfo.AsyncResult.IsBusy			= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
		
		return ret;
	}
	
	return SOCKET_NOERR_ASYNCH_RUNNING;
}



int SocketReceiveMacRaw(Socket_Type* socket, uint8_t* buffer, uint16_t bufferSize, uint16_t* receivedBytes)
{
	*receivedBytes = 0;
	
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;

	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxRxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXRXB(socket->Channel);

	uint16_t rxSize;
	CHECK_RETV_RETV(GetSxReceivedSize(SxReg, &rxSize));
	if (rxSize == 0) { return SOCKET_NOERR; }

	uint16_t rxRd;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, &rxRd));
	
	uint16_t dataLength;
	CHECK_RETV_RETV(WNetReadHWord(rxRd, SxRxB, &dataLength));
	dataLength	+= -2;
	rxRd		+=  2;
	
	if (dataLength > bufferSize)
	{
		CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + dataLength));

		CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));

		return SOCKET_ERROR_RBUFFER_SIZE;
	}
	
	CHECK_RETV_RETV(WNetReadBytes(rxRd, SxRxB, buffer, dataLength));
	
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_RX_RD_ADDRESS, SxReg, rxRd + dataLength));

	CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_RECV));
	
	*receivedBytes = dataLength;
	
	return SOCKET_NOERR;
}


// ----------------------------------------------------------------------------------------------


int SocketBeginSendMacRaw(Socket_Type* socket, uint8_t* buffer, uint16_t count, SocketAsyncCallBack_Type callBack)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;
		
	SxAsyncInfo.AsyncResult.IsCompleted		= false;
	SxAsyncInfo.AsyncResult.IsBusy			= false;
	SxAsyncInfo.AsyncResult.AsyncState		= WNET_NOERR;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
	SxAsyncInfo.AsyncResult.Socket = 0;
		
	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxTxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXTXB(socket->Channel);

	uint16_t txFreeSize;
	CHECK_RETV_RETV(GetSxTxBufferFreeSize(SxReg, &txFreeSize));
	if (txFreeSize < count) { return SOCKET_ERROR_TBUFFER_SIZE; }

	uint16_t txWr;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, &txWr));
	
	SxAsyncInfo.ScrCmd		= WNET_SRB_SCR_CMD_SEND;
	SxAsyncInfo.TxWr		= txWr + count;
	SxAsyncInfo.CallBack	= callBack;
	SxAsyncInfo.IsActive	= true;

	SxAsyncInfo.AsyncResult.IsBusy			= true;
	SxAsyncInfo.AsyncResult.ReadWriteBytes	= count;
	SxAsyncInfo.AsyncResult.Socket			= socket;
	
	int ret = WNetBeginWriteBytes(txWr, SxTxB, buffer, count, AsyncCallBackWrite);
	if (ret)
	{
		SxAsyncInfo.IsActive	= false;
		
		SxAsyncInfo.AsyncResult.IsBusy			= false;
		SxAsyncInfo.AsyncResult.ReadWriteBytes	= 0;
		
		return ret;
	}
			
	return SOCKET_NOERR_ASYNCH_RUNNING;
}



int SocketSendMacRaw(Socket_Type* socket, uint8_t* buffer, uint16_t count)
{
	if (SxAsyncInfo.IsActive)
		return SOCKET_ERROR_ASYNCH_BUSY;

	uint8_t SxReg = WNET_FRAME_CONTROL_PHASE_BSEL_SXREG(socket->Channel);
	uint8_t SxTxB = WNET_FRAME_CONTROL_PHASE_BSEL_SXTXB(socket->Channel);

	uint16_t txFreeSize;
	CHECK_RETV_RETV(GetSxTxBufferFreeSize(SxReg, &txFreeSize));
	if (txFreeSize < count) { return SOCKET_ERROR_TBUFFER_SIZE; }

	uint16_t txWr;
	CHECK_RETV_RETV(WNetReadHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, &txWr));
	CHECK_RETV_RETV(WNetWriteBytes(txWr, SxTxB, buffer, count));

	CHECK_RETV_RETV(WaitUntilSxLastSendFinished(SxReg));
	
	CHECK_RETV_RETV(WNetWriteHWord(WNET_SRB_TX_WR_ADDRESS, SxReg, txWr + count));

	CHECK_RETV_RETV(WNetWriteByte(WNET_SRB_SCR_ADDRESS, SxReg, WNET_SRB_SCR_CMD_SEND));

	StatusLedAddOneFlash();

	return SOCKET_NOERR;
}
